/* $Id$ */
/*
 * JFIF/JPEG decompression.
 */
#include <stdlib.h>
#include <stdio.h>
#include <setjmp.h>
#include <jpeglib.h>
#include "module.h"
#if defined ENABLE_CMS
#include "iccjpeg.h"
#include "cms.h"
#endif /* ENABLE_CMS */

struct my_error_mgr {
  struct jpeg_error_mgr pub;	/* "public" fields */
  jmp_buf setjmp_buffer;	/* for return to caller */
};
typedef struct my_error_mgr *my_error_ptr;

int mod_read_JPEG(Entry *entry, ImageSet *images);
static void set_colormap(struct jpeg_decompress_struct *cinfo, Image *image);
#if defined ENABLE_CMS
static CmsProfile *get_cms_profile(j_decompress_ptr cinfo);
#endif /* ENABLE_CMS */
static void my_error_exit(j_common_ptr cinfo);
static void my_error_output(j_common_ptr cinfo);
#if defined DEBUG
static void info_print(struct jpeg_decompress_struct *cinfo);
#endif /* DEBUG */

int mod_read_JPEG(Entry *entry, ImageSet *images)
{
  struct jpeg_decompress_struct cinfo;
  struct my_error_mgr jerr;
  FILE * fp;
  Image *image;

  image = image_new();
  if (image == NULL) {
    return FALSE;
  }
  if ((fp = fopen(entry->filename, "rb")) == NULL) {
    fprintf(stderr, "can't open %s\n", entry->filename);
    return FALSE;
  }
  /* We set up the normal JPEG error routines, then override error_exit. */
  cinfo.err = jpeg_std_error(&jerr.pub);
  jerr.pub.error_exit = my_error_exit;
  jerr.pub.output_message = my_error_output;
  if (setjmp(jerr.setjmp_buffer)) {
    fclose(fp);
    jpeg_destroy_decompress(&cinfo);
    return FALSE;
  }
  /* Now we can initialize the JPEG decompression object. */
  jpeg_create_decompress(&cinfo);
  jpeg_stdio_src(&cinfo, fp);
  (void)jpeg_read_header(&cinfo, TRUE);

  jpeg_calc_output_dimensions(&cinfo);
  int w = cinfo.output_width;
  int h = cinfo.output_height;
  int bperpix = cinfo.output_components;
  int bperline = w*bperpix;
  int count = h*bperline;
  cinfo.dct_method = JDCT_ISLOW; /* JDCT_ISLOW; JDCT_IFAST; JDCT_FLOAT; */
  if (w <= 0 || h <= 0 || bperline/w < bperpix || count/h < bperline) {
    jpeg_destroy_decompress(&cinfo);
    fclose(fp);
    return FALSE;
  }
  image_set_type(image, (bperpix == 1) ? IMAGE_TYPE_INDEXED : IMAGE_TYPE_RGB_8);
  set_colormap(&cinfo, image);
  unsigned char *data = (unsigned char *)malloc(count);
  if (data == NULL) {
    jpeg_destroy_decompress(&cinfo);
    fclose(fp);
    return FALSE;
  }

  (void)jpeg_start_decompress(&cinfo);
  while (cinfo.output_scanline < cinfo.output_height) {
    if (cinfo.output_scanline < 0) {
      return FALSE;
    }
    JSAMPROW rowptr[1];
    rowptr[0] = (JSAMPROW) &data[cinfo.output_scanline*bperline];
    (void)jpeg_read_scanlines(&cinfo, rowptr, (JDIMENSION)1);
  }
  (void)jpeg_finish_decompress(&cinfo);

  image_set_data(image, data);
  image_set_size(image, w, h);

  imageSet_set_size(images, w, h);
  imageSet_add_image(images, image);
#if defined DEBUG
  fprintf(stderr, "jpeg: [%dx%d], bpl = %d, bpp = %d\n",
          w, h, bperline, bperpix);
  info_print(&cinfo);
#endif /* DEBUG */
  jpeg_destroy_decompress(&cinfo);
  fclose(fp);
  return TRUE;
}

static void set_colormap(struct jpeg_decompress_struct *cinfo, Image *image)
{
  if (cinfo->out_color_space == JCS_GRAYSCALE) {
    int n = 256;
    image_set_num_colors(image, n);
    for (int i = 0; i < n; i++) {
      image_set_color(image, i, i, i, i, 0xff);
    }
  } else {
    if (cinfo->quantize_colors) {
      int n = cinfo->actual_number_of_colors;
      image_set_num_colors(image, n);
      for (int i = 0; i < n; i++) {
        image_set_color(image, i, cinfo->colormap[0][i], cinfo->colormap[1][i], cinfo->colormap[2][i], 0xff);
      }
    }
  }
}

static void my_error_exit(j_common_ptr cinfo)
{
  my_error_ptr myerr;

  myerr = (my_error_ptr)cinfo->err;
  (*cinfo->err->output_message)(cinfo);
  longjmp(myerr->setjmp_buffer, 1);
}

static void my_error_output(j_common_ptr cinfo)
{
  my_error_ptr myerr;
  char buffer[JMSG_LENGTH_MAX];

  myerr = (my_error_ptr)cinfo->err;
  (*cinfo->err->format_message)(cinfo, buffer);
  fprintf(stderr, "%s\n", buffer);
}

#if defined ENABLE_CMS
/*
 * This snippet shows how to deal with Embedded profiles in JPEG
 * files using IJG (aka, Tom Lane's) library. You will also need
 * the icc decoder, present on IJG site.
 */
static CmsProfile *get_cms_profile(j_decompress_ptr cinfo)
{
  CmsProfile *profile;
  JOCTET buffer;
  unsigned int len;

  if (read_icc_profile(cinfo, &buffer, &len)) {
    profile = cmsProfile_open_mem(buffer, len);
    XFREE(buffer);
    return profile;
  }
  return NULL;
}
#endif /* ENABLE_CMS */

#if defined DEBUG
static void info_print(struct jpeg_decompress_struct *cinfo)
{
  J_COLOR_SPACE out_color_space;
  char *type;
  out_color_space = cinfo->out_color_space;
  switch (out_color_space) {
  case JCS_UNKNOWN:
    type = "unknown";
    break;
  case JCS_GRAYSCALE:
    type = "grayscale";
    break;
  case JCS_RGB:
    type = "RGB";
    break;
  case JCS_YCbCr:
    type = "YCbCr";
    break;
  case JCS_CMYK:
    type = "CMYK";
    break;
  case JCS_YCCK:
    type = "YCbCrK";
    break;
  }
  fprintf(stderr, "jpeg: %s\n", type);
}
#endif /* DEBUG */
