/* $Id$ */
/*
 * PNG decoding.
 */
#include <stdlib.h>
#include <stdio.h>
#include <limits.h>
#include <png.h>
#include <setjmp.h>
#include "module.h"
#if defined ENABLE_CMS
#include "cms.h"
#endif /* ENABLE_CMS */

int mod_read_PNG(Entry *entry, ImageSet *images);
static int init_png_create(png_structp *png_ptr, png_infop *info_ptr);
static int get_png(png_structp png_ptr, png_infop info_ptr, ImageSet *images, Entry *entry);
static int update_info(png_structp png_ptr, png_infop info_ptr, ImageSet *images);
static Image *get_image(png_structp png_ptr, png_infop info_ptr,
                        png_uint_32 width, png_uint_32 height,
                        int pass);
static int set_imagetype(png_structp png_ptr, png_infop info_ptr,
                         png_uint_32 width, png_uint_32 height, int color_type,
                         Image **image_ptr, int *linesize_ptr);
static const char *get_comment(png_structp png_ptr, png_infop info_ptr);
static void my_error(png_structp png_ptr, png_const_charp message);
static void my_warning(png_structp png_ptr, png_const_charp message);
#if defined PNG_APNG_SUPPORTED
static int get_apng(png_structp png_ptr, png_infop info_ptr, ImageSet *images, Entry *entry);
static int get_dispose_op(png_byte dispose, png_uint_32 index);
static int get_blend_op(png_byte blend);
static int get_delay(png_uint_16 next_frame_delay_num, png_uint_16 next_frame_delay_den);
#endif /* PNG_APNG_SUPPORTED */
#if defined ENABLE_CMS
static CmsProfile *get_cms_profile(png_structp png_ptr, png_infop info_ptr);
#endif /* ENABLE_CMS */
#if defined DEBUG
static void info_print(png_infop info_ptr);
#endif /* DEBUG */


int mod_read_PNG(Entry *entry, ImageSet *images)
{
  png_structp png_ptr = NULL;
  png_infop info_ptr = NULL;
  FILE *fp;
  int rc = FALSE;

  if ((fp = fopen(entry->filename, "rb")) == NULL) {
    fprintf(stderr, "can't open %s\n", entry->filename);
    return FALSE;
  }
  if (!init_png_create(&png_ptr, &info_ptr)) {
    return FALSE;
  }
  /* setjmp() must be called in every function that calls a PNG-reading
   * libpng function */
  if (setjmp(png_jmpbuf(png_ptr))) {
    png_destroy_read_struct(&png_ptr, &info_ptr, png_infopp_NULL);
    fclose(fp);
    return FALSE;
  }
  /* I/O initialization */
  png_init_io(png_ptr, fp);

  png_read_info(png_ptr, info_ptr);  /* read all PNG info up to image data */
#if defined PNG_APNG_SUPPORTED
  if (png_get_valid(png_ptr, info_ptr, PNG_INFO_acTL)) {
    rc = get_apng(png_ptr, info_ptr, images, entry);
  } else {
    rc = get_png(png_ptr, info_ptr, images, entry);
  }
#else
  rc = get_png(png_ptr, info_ptr, images, entry);
#endif /* PNG_APNG_SUPPORTED */
  png_read_end(png_ptr, info_ptr);

  /* finish */
  png_destroy_read_struct(&png_ptr, &info_ptr, png_infopp_NULL);
  fclose(fp);
  return rc;
}

static int init_png_create(png_structp *png_ptr, png_infop *info_ptr)
{
  *png_ptr = png_create_read_struct(PNG_LIBPNG_VER_STRING, png_voidp_NULL, my_error, my_warning);
  if (*png_ptr == NULL) {
    return FALSE;
  }
  *info_ptr = png_create_info_struct(*png_ptr);
  if (*info_ptr == NULL) {
    png_destroy_read_struct(png_ptr, png_infopp_NULL, png_infopp_NULL);
    return FALSE;
  }
  return TRUE;
}

#if defined PNG_APNG_SUPPORTED
static int get_apng(png_structp png_ptr, png_infop info_ptr, ImageSet *images, Entry *entry)
{
  png_uint_32 num_frames;
  png_uint_32 num_plays;
  png_uint_32 width, height;
  int bit_depth, color_type;

  num_frames = png_get_num_frames(png_ptr, info_ptr);
  num_plays = png_get_num_plays(png_ptr, info_ptr);
  png_get_IHDR(png_ptr, info_ptr, &width, &height, &bit_depth, &color_type,
               NULL, NULL, NULL);
#if defined DEBUG
  fprintf(stderr, "png: APNG, num_frames=%d, num_plays=%d.\n",
          num_frames, num_plays);
  info_print(info_ptr);
#endif /* DEBUG */
  int pass = update_info(png_ptr, info_ptr, images);
#if defined DEBUG
  info_print(info_ptr);
#endif /* DEBUG */
  imageSet_set_size(images, width, height);
  imageSet_set_num_plays(images, num_plays);
  for (png_uint_32 i = 0; i < num_frames; i++) {
    png_uint_32 next_frame_width;
    png_uint_32 next_frame_height;
    png_uint_32 next_frame_x_offset = 0;
    png_uint_32 next_frame_y_offset = 0;
    int delay = 0;
    int dispose_op = DISPOSE_OP_NONE;
    int blend_op = BLEND_OP_NONE;
    Image *image;

    png_read_frame_head(png_ptr, info_ptr);
    if (png_get_valid(png_ptr, info_ptr, PNG_INFO_fcTL)) {
      png_uint_16 next_frame_delay_num;
      png_uint_16 next_frame_delay_den;
      png_byte next_frame_dispose_op;
      png_byte next_frame_blend_op;
      png_get_next_frame_fcTL(png_ptr, info_ptr,
                              &next_frame_width, &next_frame_height,
                              &next_frame_x_offset, &next_frame_y_offset,
                              &next_frame_delay_num, &next_frame_delay_den,
                              &next_frame_dispose_op, &next_frame_blend_op);
      delay = get_delay(next_frame_delay_num, next_frame_delay_den);
      dispose_op = get_dispose_op(next_frame_dispose_op, i);
      blend_op = get_blend_op(next_frame_blend_op);
    } else {
      /* the first frame doesn't have an fcTL so it's expected to be hidden, 
       * but we'll extract it anyway */
      next_frame_width = png_get_image_width(png_ptr, info_ptr);
      next_frame_height = png_get_image_height(png_ptr, info_ptr);
    }
    image = get_image(png_ptr, info_ptr, next_frame_width, next_frame_height, pass);
    if (image == NULL) {
      return FALSE;
    }
    image_set_delay(image, delay);
    image_set_dispose_op(image, dispose_op);
    image_set_blend_op(image, blend_op);
    image_set_offset(image, next_frame_x_offset, next_frame_y_offset);
    char *comment;
    comment = get_comment(png_ptr, info_ptr);
    image_set_comment(image, comment);

    imageSet_add_image(images, image);
  }
  return TRUE;
}
#endif /* PNG_APNG_SUPPORTED */

static int get_png(png_structp png_ptr, png_infop info_ptr, ImageSet *images, Entry *entry)
{
  png_uint_32 width, height;
  int bit_depth, color_type;
  Image *image;
  const char *comment;

  png_get_IHDR(png_ptr, info_ptr, &width, &height, &bit_depth, &color_type,
               NULL, NULL, NULL);
#if defined DEBUG
  info_print(info_ptr);
#endif /* DEBUG */
  int pass = update_info(png_ptr, info_ptr, images);
#if defined DEBUG
  info_print(info_ptr);
#endif /* DEBUG */
  imageSet_set_size(images, width, height);

  image = get_image(png_ptr, info_ptr, width, height, pass);
  if (image == NULL) {
    return FALSE;
  }
  comment = get_comment(png_ptr, info_ptr);
  image_set_comment(image, comment);
  imageSet_add_image(images, image);
  return TRUE;
}

static int update_info(png_structp png_ptr, png_infop info_ptr, ImageSet *images)
{
  png_byte color_type;
  png_byte bit_depth;

  color_type = png_get_color_type(png_ptr, info_ptr);
  bit_depth = png_get_bit_depth(png_ptr, info_ptr);
  if (color_type == PNG_COLOR_TYPE_GRAY_ALPHA) {
    png_set_gray_to_rgb(png_ptr);
  }
  if (png_get_valid(png_ptr, info_ptr, PNG_INFO_tRNS)) {
    /* transform transparency to alpha */
    switch (color_type) {
    case PNG_COLOR_TYPE_PALETTE:
      break;
    case PNG_COLOR_TYPE_GRAY:
      if (bit_depth > 8) {
        png_set_tRNS_to_alpha(png_ptr);
      }
      break;
    default:
      png_set_tRNS_to_alpha(png_ptr);
      break;
    }
  }
  if (bit_depth == 16) {
    /* reduce 16bit/channel to 8bit/channel */
    png_set_strip_16(png_ptr);
  }
  if (bit_depth < 8) {
    png_set_packing(png_ptr);
  }
  int pass = png_set_interlace_handling(png_ptr);
  /* reread infostruct to reflect the made settings */
  png_read_update_info(png_ptr, info_ptr);

  if (png_get_valid(png_ptr, info_ptr, PNG_INFO_bKGD)) {
    png_color_16p background;
    png_uint_16 red, blue, green;
    png_byte index;
    png_get_bKGD(png_ptr, info_ptr, &background);
    color_type = png_get_color_type(png_ptr, info_ptr);
    switch (color_type) {
    case PNG_COLOR_TYPE_PALETTE:
      index = background->index;
      imageSet_set_background_index(images, index);
      break;
    case PNG_COLOR_TYPE_GRAY:
      index = (background->gray >> 8) & 0xff;
      imageSet_set_background_index(images, index);
      break;
    case PNG_COLOR_TYPE_GRAY_ALPHA:
      fprintf(stderr, "Error: updated color_type gray+alpha.\n");
      break;
    case PNG_COLOR_TYPE_RGB:
    case PNG_COLOR_TYPE_RGB_ALPHA:
      bit_depth = png_get_bit_depth(png_ptr, info_ptr);
      if (bit_depth > 8) {
        red = background->red;
        green = background->green;
        blue = background->blue;
      } else {
        red = (background->red << 8) & 0xff00;
        red |= background->red & 0xff;
        green = (background->green << 8) & 0xff00;
        green |= background->green & 0xff;
        blue = (background->blue << 8) & 0xff00;
        blue |= background->blue & 0xff;
      }
      imageSet_set_background_rgb(images, red, green, blue);
      break;
    }
  }
  return pass;
}

static Image *get_image(png_structp png_ptr, png_infop info_ptr,
                        png_uint_32 width, png_uint_32 height,
                        int pass)
{
  Image *image;

  /* set image type and buffer */
  int linesize;
  int color_type = png_get_color_type(png_ptr, info_ptr);
  if (!set_imagetype(png_ptr, info_ptr, width, height, color_type, &image, &linesize)) {
    return NULL;
  }

  /* read png image */
  png_start_read_image(png_ptr);
  for (int j = 0; j < pass; j++) {
    unsigned char *p = image_get_data(image);
    for (int i = 0; i < height; i++) {
      png_read_row(png_ptr, p, png_bytep_NULL);
      p += linesize;
    }
  }
  return image;
}

static int set_imagetype(png_structp png_ptr, png_infop info_ptr,
                         png_uint_32 width, png_uint_32 height, int color_type,
                         Image **image_ptr, int *linesize_ptr)
{
  Image *image;
  ImageType type;
  unsigned char *data;
  int bufsize, linesize;
  png_colorp palette;
  int num_palette;
  int bit_depth;

  image = image_new();
  if (image == NULL) {
    return FALSE;
  }
  image_set_size(image, width, height);
  bit_depth = png_get_bit_depth(png_ptr, info_ptr);
  switch (color_type) {
  case PNG_COLOR_TYPE_RGB:
    linesize = bit_depth*3/CHAR_BIT*width;
    type = (bit_depth > 8) ? IMAGE_TYPE_RGB_16 : IMAGE_TYPE_RGB_8;
    break;
  case PNG_COLOR_TYPE_RGB_ALPHA:
    linesize = bit_depth*4/CHAR_BIT*width;
    type = (bit_depth > 8) ? IMAGE_TYPE_RGBA_16 : IMAGE_TYPE_RGBA_8;
    break;
  case PNG_COLOR_TYPE_GRAY:
    linesize = bit_depth/CHAR_BIT*width;
    type = IMAGE_TYPE_INDEXED;
    num_palette = (0x1 << (bit_depth-1));
    image_set_num_colors(image, num_palette);
    for (int i = 0; i < num_palette; i++) {
      image_set_color(image, i, i, i, i, 0xff);
    }
    if (png_get_valid(png_ptr, info_ptr, PNG_INFO_tRNS)) {
      png_bytep trans;
      int num_trans;
      png_color_16p trans_values;
      png_get_tRNS(png_ptr, info_ptr, &trans, &num_trans, &trans_values);

      for (int i = 0; i < num_palette; i++) {
        unsigned char r, g, b, a;
        image_get_color(image, i, &r, &g, &b, &a);
        a = (i == trans_values->gray) ? 0x0: *trans;
        image_set_color(image, i, i, i, i, a);
        trans++;
      }
    }
    break;
  case PNG_COLOR_TYPE_PALETTE:
    linesize = bit_depth/CHAR_BIT*width;
    type = IMAGE_TYPE_INDEXED;
    if (png_get_PLTE(png_ptr, info_ptr, &palette, &num_palette)) {
      image_set_num_colors(image, num_palette);
      for (int i = 0; i < num_palette; i++) {
        image_set_color(image, i, palette[i].red, palette[i].green, palette[i].blue, 0xff);
      }
    }
    if (png_get_valid(png_ptr, info_ptr, PNG_INFO_tRNS)) {
      png_bytep trans;
      int num_trans;
      png_color_16p trans_values;
      png_get_tRNS(png_ptr, info_ptr, &trans, &num_trans, &trans_values);
      for (int i = 0; i < num_trans; i++) {
        unsigned char r, g, b, a;
        image_get_color(image, i, &r, &g, &b, &a);
        a = *trans;
        image_set_color(image, i, r, g, b, a);
        trans++;
      }
    }
    break;
  default:
    fprintf(stderr, "png: color_type=%d why?\n", color_type);
    break;
  }
  bufsize = linesize*height;
  if (bufsize/linesize < height) {
    return FALSE;
  }
  data = (unsigned char *)calloc(bufsize, sizeof(unsigned char));
  if (data == NULL) {
    free(image);
    return FALSE;
  }
  image_set_type(image, type);
  image_set_data(image, data);
  *linesize_ptr = linesize;
  *image_ptr = image;
  return TRUE;
}

static const char *get_comment(png_structp png_ptr, png_infop info_ptr)
{
  char *comment = NULL;
  png_textp text = NULL;
  int num_text = 0;

  png_get_text(png_ptr, info_ptr, &text, &num_text);
  if (num_text > 0) {
    const char *sep = "::";
    const char *eot = "\n";
    size_t size = 1;
    for (int i = 0; i < num_text; i++) {
      size += strlen(text[i].key)+strlen(sep)+strlen(text[i].text)+strlen(eot);
    }
    comment = (char *)malloc(size);
    if (comment == NULL) {
      png_warning(png_ptr, "can't allocate comment string");
    } else {
      comment[0] = '\0';
      for (int i = 0; i < num_text; i++) {
        strcat(comment, text[i].key);
        strcat(comment, sep);
        strcat(comment, text[i].text);
        strcat(comment, eot);
      }
    }
  }
  return comment;
}

static void my_error(png_structp png_ptr, png_const_charp message)
{
  fprintf(stderr, "%s\n", message);
  longjmp(png_ptr->jmpbuf, 1);
}

static void my_warning(png_structp png_ptr, png_const_charp message)
{
  fprintf(stderr, "%s\n", message);
}

#if defined PNG_APNG_SUPPORTED
static int get_dispose_op(png_byte dispose, png_uint_32 index)
{
  int dispose_op = DISPOSE_OP_NONE;
  switch (dispose) {
  case PNG_DISPOSE_OP_NONE:
    dispose_op = DISPOSE_OP_NONE;
    break;
  case PNG_DISPOSE_OP_BACKGROUND:
    dispose_op = DISPOSE_OP_BACKGROUND;
    break;
  case PNG_DISPOSE_OP_PREVIOUS:
    dispose_op = (index != 0) ? DISPOSE_OP_PREVIOUS : DISPOSE_OP_BACKGROUND;
    break;
  }
  return dispose_op;
}

static int get_blend_op(png_byte blend)
{
  int blend_op = BLEND_OP_SOURCE;
  switch (blend) {
  case PNG_BLEND_OP_SOURCE:
    blend_op = BLEND_OP_SOURCE;
    break;
  case PNG_BLEND_OP_OVER:
    blend_op = BLEND_OP_OVER;
    break;
  }
  return blend_op;
}

static int get_delay(png_uint_16 next_frame_delay_num, png_uint_16 next_frame_delay_den)
{
  int delay;
  if (next_frame_delay_den == 0) {
    next_frame_delay_den = 100;
  }
  delay = next_frame_delay_num*1000/next_frame_delay_den;
  return delay;
}
#endif /* PNG_APNG_SUPPORTED */

#if defined ENABLE_CMS
static CmsProfile *get_cms_profile(png_structp png_ptr, png_infop info_ptr)
{
  /* First try to see if iCCP chunk is present */
  if (png_get_valid(png_ptr, info_ptr, PNG_INFO_iCCP)) {
    png_charp *name;
    int compression;
    png_charp *profile;
    png_uint_32 len;

    png_get_iCCP(png_ptr, info_ptr, &name, &compression, &profile, &len);
    return cmsProfile_open_mem(profile, len);
  }
  /* Try to see if sRGB is specified, then use it */
  if (png_get_valid(png_ptr, info_ptr, PNG_INFO_sRGB)) {
    int intent;
    png_get_sRGB(png_ptr, info_ptr, &intent);
    return cmsProfile_open_sRGB(intent);
  }
  /* Then, try to see if chromacities are defined. */
  if (png_get_valid(png_ptr, info_ptr, PNG_INFO_cHRM) &&
      png_get_valid(png_ptr, info_ptr, PNG_INFO_gAMA)) {
    double white_x, white_y;
    double red_x, red_y;
    double green_x, green_y;
    double blue_x, blue_y;
    double gamma;

    png_get_cHRM(png_ptr, info_ptr,
                 &white_x, &white_y, &red_x, &red_y,
                 &green_x, &green_y, &blue_x, &blue_y);
    png_get_gAMA(png_ptr, info_ptr, &gamma);
    return cmsProfile_open_chromaticities(white_x, white_y, red_x, red_y,
                                          green_x, green_y, blue_x, blue_y,
                                          gamma);
  }
  /* Only gamma is given, assume sRGB primaries?? */
  if (png_get_valid(png_ptr, info_ptr, PNG_INFO_gAMA)) {
    double gamma;

    png_get_gAMA(png_ptr, info_ptr, &gamma);
    return cmsProfile_open_D65(gamma);
  }
  /* The image cannot be accurately described */
  return NULL;
}
#endif /* ENABLE_CMS */

#if defined DEBUG
static void info_print(png_infop info_ptr)
{
  char *type;
  switch(info_ptr->color_type) {
  case PNG_COLOR_TYPE_PALETTE:
    type = "palette";
    break;
  case PNG_COLOR_TYPE_GRAY:
    type = "gray";
    break;
  case PNG_COLOR_TYPE_GRAY_ALPHA:
    type = "gray+alpha";
    break;
  case PNG_COLOR_TYPE_RGB:
    type = "rgb";
    break;
  case PNG_COLOR_TYPE_RGB_ALPHA:
    type = "rgb+alpha";
    break;
  }
  fprintf(stderr, "png: (%dx%d) %d bit %s [%dx%d]\n",
          info_ptr->bit_depth, info_ptr->channels,
          info_ptr->bit_depth*info_ptr->channels,
          type, (int)info_ptr->width, (int)info_ptr->height);
}
#endif /* DEBUG */
